// FROG_Element_Base.h: Contains everything related to an element to be written in a file :
//                     type, size, data
//                     And a vector of daughters
//////////////////////////////////////////////////////////////////////

#ifndef _FROG_ELEMENT_BASE_H__
#define _FROG_ELEMENT_BASE_H__

#include "FROG_Chunk.h"
#include "FROG_Objects_Style.h"
#include "FROG_Light.h"

#ifdef FROG_OPENGL
	#ifdef linux
		#include <GL/gl.h>
		#include <GL/glu.h> 
		#include "../GLUT/glut.h"
    #elif macos
            #include <OpenGL/gl.h>
            #include <OpenGL/glu.h>
            #include <GLUT/glut.h>
	#else		
		#include <windows.h> 
		#include <gl/gl.h>
		#include <gl/glu.h>
		#include <gl/glut.h>
	#endif
#endif


#include <vector>
#include <map>
#include <algorithm>
#include <math.h>

class FROG_Element_Base{
protected :
	unsigned short type_;
	unsigned int   size_;
	unsigned char* data_;
	

	std::vector<FROG_Element_Base*> daughters_;
	FROG_Element_Base* mother_;
	
	unsigned char display_;
	unsigned int  display_list_;

public:
	FROG_Objects_Style* style_;	
	bool MenuOpened_;

public :
	class cmp {
	public:
		bool operator () (FROG_Element_Base* a, FROG_Element_Base* b) {return (a->type() < b->type());}
	};

	FROG_Element_Base(unsigned short type) : type_(type), size_(6), data_(NULL), mother_(NULL), display_(0), display_list_(0), style_(NULL), MenuOpened_(false){}

	virtual ~FROG_Element_Base(){		
		for(unsigned int i=0;i<daughters_.size();i++){if(daughters_[i]!=NULL){
			delete daughters_[i];
			daughters_[i]=NULL;}
		}
		clearData();
	}

	virtual bool        isCompactible() { return false; }  // only leaf should be compactible
	static unsigned int sizeOf()        { return 6;}

	virtual void clearData()          { if(data_)delete [] data_; data_=NULL; size_ = 6;}
	virtual void clearDaughtersData() { 
		clearData();
		for(unsigned int i=0;i<daughters_.size();i++){
			if(daughters_[i]!=NULL)daughters_[i]->clearDaughtersData();
		}
	}
	
    unsigned int   size() const { return size_; }
    unsigned short type() const { return type_; }
	virtual  void* data() const { return data_; }
	const std::vector<FROG_Element_Base*>& daughters() const { return daughters_; }
	void set_mother(FROG_Element_Base* mother){ mother_ = mother;}
	
    void addDaughter( FROG_Element_Base* dau){
		dau->set_mother(this);
       	daughters_.push_back(dau);
    }

    FROG_Element_Base* getDaughter(unsigned int i){		
       	return daughters_[i];
    }

	
	void GetFullDaughtersVector(std::vector<FROG_Element_Base*>* vectorToFill){
		for(unsigned int i=0;i<daughters_.size();i++){
			vectorToFill->push_back(daughters_[i]);
			daughters_[i]->GetFullDaughtersVector(vectorToFill);
			//printf("GETFULLVECTORR\n");
		}
	}

    unsigned int getDaughterSize(){
       	return daughters_.size();
    }

	void deleteDaughter(unsigned int i){
		if(i>=daughters_.size())return;

		daughters_[i]->deleteDaughters();
		delete daughters_[i];		
		daughters_.erase(daughters_.begin()+i);
	}

	void deleteDaughters(){
		for(unsigned int i=0;i<daughters_.size();i++){
			daughters_[i]->deleteDaughters();
			delete daughters_[i];
		}
		daughters_.clear();
	}

    unsigned int displayList()      { return display_list_; }


    FROG_Element_Base* getMother(){
       	return mother_;
    }

    unsigned char* FillBuffer(unsigned char* ptr, void* x, int size){
		memcpy(ptr, x, size);		
		return (unsigned char*)((unsigned long)ptr + size);
    }

	// Prepare data_ for data storing
    std::vector< std::pair<unsigned int, unsigned int > >* write_init(){
		// take care of empty collections
		if(daughters_.size()==0){
			if(size_-6>0){
				data_ = new unsigned char[size_-6];
				return NULL;
			}else{
				data_ = NULL;
				return NULL;
			}
		}
		
		// sort the daughters with respect to their type
		std::stable_sort(daughters_.begin(), daughters_.end(), cmp());

		// all the daughters are written
		for(unsigned int i=0;i<daughters_.size();i++){
			daughters_[i]->write();
		}

		// the first type in the daughter collection
		unsigned short previous_type = 0;
		unsigned int   previous_size = 0;

		// pair of type X and size of daughters of type X
		std::vector< std::pair<unsigned int, unsigned int > >* blockOfDaughters = new std::vector< std::pair<unsigned int, unsigned int > >;
		int ndau=1; // number of daughters of type X		 
		for(unsigned int i=0; i<=daughters_.size(); i++) {           
			if(i<daughters_.size() && previous_type==daughters_[i]->type() && daughters_[i]->isCompactible()){
				ndau++;
			}else{			   
				if(i!=0){					
					blockOfDaughters->push_back( std::make_pair< unsigned int, unsigned int >(i-ndau, ndau) );			   					
					size_ += (previous_size-6)*ndau;					
					//printf("write : %i forloop --> %i size_+= (%i-6)*%i --> size_ = %i\n",type_, i-1, previous_size, ndau,size_);
				}

				if(i<daughters_.size()){
					previous_type = daughters_[i]->type();
					previous_size = daughters_[i]->size();
					ndau = 1;
				}			   
			}
		}		
		size_ += 6*blockOfDaughters->size();
		
		//printf("write : %i size_+= 6*%i --> size_ = %i\n",type_, blockOfDaughters.size(), size_);				
		if(data_)delete [] data_;
		data_ = new unsigned char[size_-6];					

		return blockOfDaughters;
	}

	// Write daughter's data in data_
	void write_daughters(std::vector< std::pair<unsigned int, unsigned int > >* blockOfDaughters){
		if(blockOfDaughters==NULL){
			data_ = (unsigned char*)((unsigned long)data_ - (size_-6) );
			return;
		}

		// write the type and the size of the daughters of one type followed by their data
		for(unsigned int k=0; k<blockOfDaughters->size(); k++) {		  
			unsigned short tmp_type = daughters_[(*blockOfDaughters)[k].first]->type();
			unsigned int   tmp_size = 6+(daughters_[(*blockOfDaughters)[k].first]->size()-6)*(*blockOfDaughters)[k].second;
			data_ = FillBuffer(data_, &tmp_type, sizeof(unsigned short)); // write the type		  
			data_ = FillBuffer(data_, &tmp_size, sizeof(unsigned int)  ); // write the size

			for(unsigned int i=0; i<(*blockOfDaughters)[k].second; i++) {
				unsigned int DaughterIndex = (*blockOfDaughters)[k].first + i;
				if(daughters_[DaughterIndex]->data()==NULL)continue;				
				data_ = FillBuffer(data_, daughters_[DaughterIndex]->data(), daughters_[DaughterIndex]->size()-6 );				
				daughters_[DaughterIndex]->clearData();				
			}		
		}

		// put the pointer at the begin of the data array
		data_ = (unsigned char*)((unsigned long)data_ - (size_-6) );
		delete blockOfDaughters;
	}

    // Fill data_ and size_ with data
    virtual void write(){
		std::vector< std::pair<unsigned int, unsigned int > >* blockOfDaughters = write_init();
		if(blockOfDaughters==NULL)return;
		write_daughters(blockOfDaughters);			
	}

	unsigned char getDisplay(){	return display_; }

    void setDisplay(unsigned char flag){
		#ifdef FROG_OPENGL
			CleardisplayList();
		#endif

        unsigned int dauSize = daughters_.size();
		if(flag<2){		
			display_ = flag;
			for(unsigned int i=0;i != dauSize;i++){
				if(daughters_[i]==NULL)continue;				
				daughters_[i]->setDisplay(flag);
			}

		}else if(flag==2){
			if(dauSize==0) return;
            unsigned int firstFlag  = daughters_[0]->display_;

			display_ = 0;
			for(unsigned int i=0;i!=dauSize;i++){
                FROG_Element_Base* dau = daughters_[i];
				if(dau==NULL)continue;
				dau->setDisplay(flag);
                if(firstFlag != dau->display_)
					display_=2; 						                
			}
			if(display_==0)
				display_=firstFlag;
		}		
		return;
    }

    void toogleParentDisplay(unsigned char flag, unsigned int level){
		if( (mother_==NULL && level>=1) || level==0){
			if(getDisplay()!=1){
				setDisplay(flag);
			}else if(getDisplay()==1){
				setDisplay(0);
			}
		}
		if(mother_!=NULL && level>=1)mother_->toogleParentDisplay(flag,level-1);		
    }

	virtual void printInfos(char* buffer){sprintf(buffer,"No Infos to print");}
   
    virtual void uninit(){ 
		if(mother_ && mother_->style_ == style_)style_=NULL;		
		for(unsigned int i=0;i<daughters_.size();i++)daughters_[i]->uninit();
		if(style_){delete style_;style_=NULL;}

		#ifdef FROG_OPENGL
              if(display_list_>0) { glDeleteLists(display_list_, 1); display_list_=0; }
		#endif		
    }

	virtual void init(void* frogObjects){
        if(display_list_>0) { 
			#ifdef FROG_OPENGL
				glDeleteLists(display_list_, 1); 
				display_list_=0; 
			#endif
		}
		if(mother_ != NULL && mother_->style_ !=NULL){
			style_ = mother_->style_;
		}else{
			style_ = new FROG_Objects_Style();
		}			

		for(unsigned int i=0;i<daughters_.size();i++)
			daughters_[i]->init(frogObjects);
	}



	#ifdef FROG_OPENGL

	void CleardisplayList() { 
		if(display_list_==0)return;
		glDeleteLists(display_list_,1); 
		display_list_ = 0; 
	}

	void SuperCleardisplayList() { 
		if(display_list_>=0){
			glDeleteLists(display_list_,1); 
			display_list_ = 0; 
		}

		for(unsigned int i=0;i<daughters_.size();i++){
			daughters_[i]->SuperCleardisplayList();			
		}
	}

	bool Start_Display(bool UseDisplayList){
		if(display_list_ == 0 && UseDisplayList){			
			display_list_ = glGenLists (1);
			glNewList(display_list_, GL_COMPILE_AND_EXECUTE);
			return true;
		}else if(UseDisplayList){			
			glCallList(display_list_);			
			return false;						
		}else{
			return true;
		}
	}

	void End_Display(bool UseDisplayList){
		if(UseDisplayList){			
			glEndList();
		}
	}


	virtual void display(bool UseDisplayList=true, float* color=NULL){
		if(display_==0)return;

		if(UseDisplayList){

			bool UnInitDaughter = false;
			for(unsigned int i=0;i<daughters_.size();i++){			
				if(daughters_[i] && (daughters_[i]->getDisplay()!= 1 || daughters_[i]->display_list_==0)){					
					UnInitDaughter = true;				
				}
			}	

			if(UnInitDaughter){
				for(unsigned int i=0;i<daughters_.size();i++){			
					if(daughters_[i] && daughters_[i]->getDisplay()> 0){
						daughters_[i]->display(1);								
					}
				}	
			}else if(display_list_==0){
				display_list_ = glGenLists (1);
				glNewList(display_list_, GL_COMPILE_AND_EXECUTE);
				for(unsigned int i=0;i<daughters_.size();i++){
					if(daughters_[i]!=NULL)daughters_[i]->display(true);
				}
				glEndList();		
			}else{				
				glCallList(display_list_);
			}
		}else{
			for(unsigned int i=0;i<daughters_.size();i++){			
				if(daughters_[i] && daughters_[i]->getDisplay()>0)	daughters_[i]->display(false);
			}	
		}

		
	}
	#else
	virtual void display(bool UseDisplayList=true, float* color=NULL){return;}
    #endif

	virtual double getPosX(){	return -1E100;	}
	virtual double getPosY(){	return -1E100;	}
	virtual double getPosZ(){	return -1E100;	}
};
#endif
